/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: LeaderCrashStopServer.java,v 1.0   Exp $
  *
  */

package universal.leader;
/*
 * Author       : Isabelle Ravot
 * Date         : 01/08/02
 * Last changes : 01/08/02
 * Contact		: diasz@ida.liu.se
 */


import universal.communication.*;
import universal.*;
import java.io.*;
import java.net.*;

/**
 *This is the implementation of the leader part of the algorithm in the server side.
 *As it is now it can be use also for the crash recovery model without big changes.
 */
public class LeaderCrashStopServer extends Leader{
    /**trustList[k] contains information about the replicated server of pid k */
    public trustListElement trustList[];

    protected class trustListElement
    {
	/** True if the replicated server k is consider to be alive*/
	protected boolean trusted;
	/**Useful only for the crash recovery model*/
	protected int epoch;
	/**The last time we have received a message AmAlive for the replicated server k*/
	protected long lastReceivingTime;
	/**
	 *If we don't have received any AmAlive message from the replicated server k during a period delta,
	 *this server is considered as not to be alive.
	 */
	protected long delta;
	/**Initialisation of all the variables. At the beginning all the replicated servers are considered to be alive.*/
	public trustListElement()
	{
	    trusted = true;
	    epoch = 0;
	    lastReceivingTime = System.currentTimeMillis();
	    delta = Constants.INITIAL_MAX_DELTA;
	}
    }
    public LeaderCrashStopServer(Universal uni)
    {
	super(uni);

	initialize(false);

    }
    /**
     *Create one object trustlist per replicated server.
     *Starts the thread.
     */
    public void initialize(boolean recovering)
    {
	trustList = new trustListElement[Constants.MAX_PROCESSES];
	for (int i=0; i<trustList.length; i++)
	    {
		trustList[i] = new trustListElement();
	    }

	Thread t = new Thread(this, "Leader");

	t.start();
    }
    /**
     * Returns first correct replicated server alive in the list.
     * As we are in the server side the function is_leader will return the pid of the leader.
     * A replicated server always trust himself
     */
    public int is_leader()
    {
	int minEpoch = Integer.MAX_VALUE;
	int leader=sender().totalno();

	for (int i=sender().totalno()-1; i>=0; i--)
	    {

		int pid = ((ProcessDetails)(sender().processesReplicas.elementAt(i))).pid;


		if (trustList[pid].trusted)
		    {

			if(pid<leader)
			    {
				leader = pid;

			    }
		    }
	    }

	Debug.println("Leader is :"+leader);
	return leader;

    }
    /**
     *Sends periodically AmAliveMessage to all the others replicated server and also to the client.
     *Check if some replicated server shouldn't be trusted anymore because we don't have receive any
     *AmAlive message from it since a period of length delta.
     */
    public synchronized void omega()
    {
	try
	    {
		Thread.currentThread().sleep(Constants.OMEGA_DELTA);
	    }
	catch (InterruptedException e)
	    {
		e.printStackTrace();
		System.err.println("Interrupted exception in omega");
	    }

	// broadcast AmAlive
	int epochArray[] = new int[Constants.MAX_PROCESSES];
	for (int i=0; i<trustList.length; i++)
	    {
		epochArray[i] = trustList[i].epoch;
	    }
	AmAliveMessage amAliveMessage = new AmAliveMessage(epochArray);

	sender().send(amAliveMessage,sender().processes);

	// update trustList
	for (int i=0; i<sender().processesReplicas.size(); i++)
	    {
		int id = ((ProcessDetails)(sender().processesReplicas.elementAt(i))).pid;
		if (trustList[id].lastReceivingTime < System.currentTimeMillis() - trustList[id].delta)
		    {
			trustList[id].trusted = false;
		    }
	    }
	// always trust myself  (not true for unstables...)
	trustList[sender().pid].trusted = true;
	//useful for debugging
	is_leader();
    }
    /**
     *Call when it receives an AmAliveMessage from one of the replicated servers.
     *If this replicated servers wasn't trusted as alive it is again trusted and we
     *increase the period delta to check if it has received a message or not.
     */
    public void receivedAmAlive(AmAliveMessage m)
    {
	Debug.println("the message received was from"+m.senderid);

	while (sender() == null)
	    {
		try{
		    Thread.currentThread().sleep(100);
		}
		catch (InterruptedException e)
		    {
			e.printStackTrace();
		    }
	    }

	trustList[m.senderid].lastReceivingTime = System.currentTimeMillis();

	if (!trustList[m.senderid].trusted)
	    {
		trustList[m.senderid].trusted = true;
		trustList[m.senderid].delta += 500;
	    }

	for (int i=0; i<m.epochArray.length; i++)
	    {
		if (trustList[i].epoch < m.epochArray[i])
		    {
			//	Debug.println("Updating "+tocs.sender.pid+"'s epoch number for "+i+" from "+trustList[i].epoch+" to "+ m.epochArray[i]);
			trustList[i].epoch = m.epochArray[i];
		    }
	    }

    }
    /*
      public void receivedRecover(RecoveredMessage m)
      {
      trustList[m.pid].lastReceivingTime = System.currentTimeMillis();

      trustList[m.pid].trusted = true;
      trustList[m.pid].epoch += 1;


}
*/
}
